@@ -190,7 +190,7 @@ ssh_connect (SshHandle **handle_return,
 
 	username = gnome_vfs_uri_get_user_name(uri);
 
-	cmd_line = g_string_new ("ssh -oBatchMode=yes -x");
+	cmd_line = g_string_new (SSH_PROGRAM " -oBatchMode=yes -x");
 	if (username != NULL) {
 		g_string_append_printf (cmd_line, " -l %s", username);
 	}
@@ -712,6 +712,10 @@ do_read_directory (GnomeVFSMethod *method,
 			get_access_info (handle->uri, file_info);
 		}
 
+		/* Use a larger block size for remote access */
+		file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_IO_BLOCK_SIZE;
+		file_info->io_block_size = 32*1024;
+		
 		/* Break out.
 		   We are in a loop so we get the first 'ls' line;
 		   often it starts with 'total 2213' etc.
